#!/usr/bin/env python3
from flask import Flask
from flask_restplus import Api, Resource, fields, abort

from . import models
from .db import Database, ProjectNotFoundError, TaskNotFoundError, JobNotFoundError, ForbiddenOperationError
from .scheduler import submit_training_job, submit_prediction_job, get_job_status, cancel_job, JobStatusNotFoundError, JobNotFoundError, JobSubmissionError

app = Flask(__name__)
# Setup flask with CORS, since we do requests to atlas controller
api = Api(app,
          prefix="/api",
          version="1.0",
          title="ATLaS UI API",
          description="API serving as backend for the ATLaS UI")

# -----------------------------------
# Namespace definitions
# -----------------------------------

project_namespace = api.namespace("projects", description="Operations to deal with projects")

# -----------------------------------
# Model definitions
# -----------------------------------

annotation_model = api.model("AnnotationModel",
                             {
                                 "section": fields.Integer(description="Section number the annotation corresponds to."),
                                 "name": fields.String(description="Name of the annotation."),
                                 "original_name": fields.String(description="Original name of the annotation before automatic renaming. May be None."),
                                 "path": fields.String(description="Path of the annotation.")
                             })

task_model = api.model("TaskModel",
                       {
                           "task_id": fields.Integer(description="ID of the task."),
                           "name": fields.String(description="The name of the task."),
                           "annotations": fields.List(fields.Nested(annotation_model), description="Annotations to add to the task."),
                       })

project_model = api.model("ProjectModel",
                          {
                              "project_id": fields.Integer(description="ID of the project."),
                              "name": fields.String(description="The name of the project."),
                              "brain": fields.Integer(description="Number of the brain the projects refers to."),
                              "owner": fields.String(description="Owner of the project."),
                              "tasks": fields.List(fields.Nested(task_model), description="List of tasks of the project.")
                          })

job_configuration_model = api.model("JobConfigurationModel",
                                    {
                                        "account": fields.String(description="Account to submit job with, e.g. jinm14.", required=True),
                                        "time": fields.String(description="Time limit for job in format HH:MM:SS.", required=True),
                                        "job_type": fields.String(description="Type of the job which is being submitted.", required=True),
                                        "gres": fields.String(description="Generic resources to request for job, e.g. \"gpus:4\""),
                                        "mail_type": fields.String(description="Mail type to use for e-mail notifications."),
                                        "mail_user": fields.String(description="E-mail address to send notifications to."),
                                        "num_nodes": fields.Integer(description="Number of nodes to allocate for the job."),
                                        "partition": fields.String(description="Partition to allocate job on, e.g. \"batch\" or \"gpus\""),
                                        "reservation": fields.String(description="Reservation to use for running the job."),
                                    })

_project_schema = models.ProjectSchema()
_task_schema = models.TaskSchema()
_job_schema = models.JobSchema()
_job_configuration_schema = models.JobConfigurationSchema()


# -----------------------------------
# Actual API definitions
# -----------------------------------

@project_namespace.route("/")
class ProjectList(Resource):
    @project_namespace.doc("Get all projects")
    def get(self):
        # TODO This function should be hidden behind an authentication mechanism
        with Database() as db:
            projects = db.projects.find({})
        if not projects:
            abort(404, f"No projects found")
        return _project_schema.dump(projects, many=True)

    @project_namespace.doc("Create a new project")
    @project_namespace.expect(project_model)
    def post(self):
        # Deserialize project
        project = _project_schema.load(api.payload)

        with Database() as db:
            project = db.insert_project(project)

        # Return the created project
        return _project_schema.dump(project)


@project_namespace.route("/<int:project_id>")
@project_namespace.response(404, "Project not found")
class Project(Resource):
    @project_namespace.doc("Get project with the specified id")
    def get(self, project_id):
        with Database() as db:
            try:
                project = db.get_project_by_id(project_id=project_id)
            except ProjectNotFoundError:
                abort(404, f"No project found for id {project_id}")
                return
        return _project_schema.dump(project)

    @project_namespace.doc("Delete project with the specified id")
    @project_namespace.response(204, "Project deleted")
    def delete(self, project_id):
        with Database() as db:
            db.delete_project(project_id=project_id)
        return {"status": "success"}

    @project_namespace.doc("Update project with the specified id")
    @project_namespace.expect(project_model)
    def put(self, project_id):
        data = api.payload
        if not data:
            abort(400, "No fields to update provided")

        for field in ("created", "modified", "project_id", "tasks"):
            data.pop(field, None)

        try:
            with Database() as db:
                db.update_project_values(project_id=project_id, key_value_dict=data)
                # Get updated project to return it
                project = db.get_project_by_id(project_id=project_id)
        except ProjectNotFoundError:
            abort(404, f"No project found for id {project_id}")
            return
        except ForbiddenOperationError as ex:
            abort(403, ex.message)
            return

        return _project_schema.dump(project)


@project_namespace.route("/users/<username>")
@project_namespace.response(404, "Project for user not found")
class UserProjectList(Resource):
    @project_namespace.doc("Get projects of the specified user")
    def get(self, username):
        with Database() as db:
            projects = db.get_projects_for_user(username=username)
        return _project_schema.dump(projects, many=True)


@project_namespace.route("/<int:project_id>/tasks")
@project_namespace.response(404, "Project not found")
class TaskList(Resource):
    @project_namespace.doc("Get tasks of the specified project")
    def get(self, project_id):
        with Database() as db:
            project = db.get_project_by_id(project_id=project_id)
        if not project:
            abort(404, f"No project found for id {project_id}")
        else:
            return _task_schema.dump(project.tasks, many=True)

    @project_namespace.doc("Create a new task for the specified project")
    @project_namespace.expect(task_model)
    def post(self, project_id):
        # Create a task
        task = _task_schema.load(api.payload)

        # Add to project
        try:
            with Database() as db:
                task = db.add_task_to_project(project_id=project_id, task=task)
        except ProjectNotFoundError:
            abort(404, f"No project found for id {project_id}")
            return

        # Return newly created task
        return _task_schema.dump(task)


@project_namespace.route("/<int:project_id>/tasks/<int:task_id>")
@project_namespace.response(404, "Task not found")
class Task(Resource):
    @project_namespace.doc("Get task with the specified task id and project id")
    def get(self, project_id, task_id):
        try:
            with Database() as db:
                task = db.get_task_by_id(project_id=project_id,
                                         task_id=task_id)
        except TaskNotFoundError:
            abort(404, f"No task found for project id {project_id} and task id {task_id}")
            return

        return _task_schema.dump(task)

    @project_namespace.doc("Delete project with the specified id")
    @project_namespace.response(204, "Task deleted")
    def delete(self, project_id, task_id):
        # Delete the task by
        try:
            with Database() as db:
                db.delete_task(project_id=project_id, task_id=task_id)
        except TaskNotFoundError:
            abort(404, f"No task found for id {task_id}")
            return
        return {"status": "success"}

    @project_namespace.doc("Update task with the specified task id and project id")
    @project_namespace.expect(task_model)
    def put(self, project_id, task_id):
        data = api.payload
        if not data:
            abort(400, "No fields to update provided")
        # Remove fields which may not be updated
        for field in ("created", "modified", "task_id", "jobs"):
            data.pop(field, None)

        try:
            with Database() as db:
                db.update_task_values(project_id=project_id, task_id=task_id, key_value_dict=data)
                # Get the modified data to return it
                task = db.get_task_by_id(project_id=project_id, task_id=task_id)
        except TaskNotFoundError:
            abort(404, f"No task found for project id {project_id} and task id {task_id}")
            return
        except ForbiddenOperationError as ex:
            abort(403, ex.message)
            return

        return _task_schema.dump(task)


@project_namespace.route("/<int:project_id>/tasks/<int:task_id>/jobs")
class TaskJobList(Resource):
    @project_namespace.doc("Get all jobs associated with a specific task")
    def get(self, project_id, task_id):
        try:
            with Database() as db:
                task = db.get_task_by_id(project_id=project_id, task_id=task_id)
        except TaskNotFoundError:
            abort(404, f"No task found for id {task_id}")
            return
        return _job_schema.dump(task.jobs, many=True)

    @project_namespace.doc("Submit a job for a specific task")
    @project_namespace.expect(job_configuration_model)
    def post(self, project_id, task_id):
        # Load the job configuration
        job_configuration = _job_configuration_schema.load(api.payload)
        # Load the model
        try:
            with Database() as db:
                project = db.get_project_by_id(project_id=project_id)
                task = db.get_task_by_id(project_id=project_id, task_id=task_id)
                brain = project.brain
                annotations = task.annotations
                # noinspection PyUnresolvedReferences
                job_type = job_configuration.job_type
                if job_type.lower() in ("training", "train"):
                    job_id = submit_training_job(brain=brain,
                                                 annotations=annotations,
                                                 job_configuration=job_configuration,
                                                 project_id=project_id,
                                                 task_id=task_id)
                elif job_type in ("prediction", "predict"):
                    job_id = submit_prediction_job(brain=brain,
                                                   job_configuration=job_configuration,
                                                   project_id=project_id,
                                                   task_id=task_id)
                else:
                    abort(400, f"Invalid job type: {job_type}")
                    return
                # Save job id into the task
                job = models.Job(job_id=job_id,
                                 job_configuration=job_configuration)
                job = _job_schema.dump(job)

                db.update_task_values(project_id=project_id,
                                      task_id=task_id,
                                      key_value_dict={"jobs": job},
                                      operation="add")

        except ProjectNotFoundError:
            abort(404, f"No project found for id {project_id}")
            return
        except TaskNotFoundError:
            abort(404, f"No task found for project id {project_id} and task id {task_id}")
            return
        except JobSubmissionError as ex:
            abort(500, f"Error submitting job: {ex.message}")
            return

        # Return serialized job
        return job


@project_namespace.route("/<int:project_id>/tasks/<int:task_id>/jobs/<int:job_id>")
class TaskJob(Resource):
    @project_namespace.doc("Get information about a specific job associated with a task")
    def get(self, project_id, task_id, job_id):
        try:
            with Database() as db:
                job = db.get_job_by_id(project_id=project_id,
                                       task_id=task_id,
                                       job_id=job_id)
        except JobNotFoundError:
            abort(404, f"No job found for project {project_id} - task {task_id} - project {project_id}")
            return

        # Dump job to dict and add the job status from the batch system
        result = _job_schema.dump(job)
        try:
            result["status"] = get_job_status(job_id=job_id)
        except JobStatusNotFoundError:
            # If the job status could not be found, it could mean the job has finished to long ago
            result["status"] = f"Could not determine job status for job with id {job_id}. " \
                f"This could mean that the job does not exist or finished too long ago."

        return result

    @project_namespace.doc("Cancel a specific job associated with a task")
    def delete(self, project_id, task_id, job_id):
        try:
            cancel_job(job_id)
            # Remove job from task
            with Database() as db:
                task = db.get_task_by_id(project_id=project_id, task_id=task_id)
                jobs = [job for job in task.jobs if job.job_id != job_id]
                jobs = _job_schema.dump(jobs, many=True)
                # Update task with changed jobs list
                db.update_task_values(project_id=project_id,
                                      task_id=task_id,
                                      key_value_dict={"jobs": jobs},
                                      operation="set")
        except JobNotFoundError:
            abort(404, f"Could not cancel job {job_id}: Job not found.")
        return {"message": "success"}


@project_namespace.route("/<int:project_id>/tasks/<int:task_id>/tile_server_config/<base_file_b64>")
class TaskTileServerConfig(Resource):

    @project_namespace.doc("Create a tile server configuration for the task")
    def get(self, project_id, task_id, base_file_b64):
        import json
        import requests
        import base64
        from atlaslib.files import get_section_number_from_string
        from .config import TILE_SERVER_OUTPUT_DIR, ATLAS_CONTROLLER_URL, TILE_SERVER_HOST

        # Download the base file
        base_file = base64.b64decode(base_file_b64).decode("utf-8")
        result = requests.get(url=base_file)
        if result.status_code >= 400:
            abort(404, f"Could not find base file {base_file}")
        try:
            base_file_content = json.loads(result.content)
        except json.JSONDecodeError:
            abort(500, f"Error decoding content of {base_file}")
            return

        try:
            with Database() as db:
                project = db.get_project_by_id(project_id=project_id)
        except ProjectNotFoundError:
                abort(404, f"No project found for id {project_id}")
                return

        try:
            with Database() as db:
                task = db.get_task_by_id(project_id=project_id, task_id=task_id)
        except ProjectNotFoundError:
                abort(404, f"No task found for id {task_id}")
                return

        # Determine
        work_dir = f"atlas_ui_project{project_id}_task{task_id}"
        # Query predictions
        result = requests.get(url=f"{ATLAS_CONTROLLER_URL}/predictions/{work_dir}")
        if result.status_code >= 400:
            abort(404, f"Failed to find predictions")
            return

        overlays = base_file_content.get("overlays", {})
        overlay_name = f"{project.name} - {task.name}"
        for prediction in result.json():
            path = "/".join([TILE_SERVER_OUTPUT_DIR, prediction["path"]])
            h, w = prediction["shape"]
            section = f"{get_section_number_from_string(path):04d}"
            section_overlays = overlays.get(section, {})
            url = "function(level, x, y){return '" + f"{TILE_SERVER_HOST}/get_tile?level=' + level + '&y=' + y + '&x=' + x + '&ts=256&img={path}&dataset=/'" + ";}"
            section_overlays[overlay_name] = {
                "getTileUrl": url,
                "height": h,
                "minLevel": 0,
                "tileOverlap": 0,
                "tileSize": 256,
                "width": w,
            }
            overlays[section] = section_overlays
        base_file_content["overlays"] = overlays

        # Return serialized job
        return base_file_content


# -----------------------------------
# Main entry point
# -----------------------------------

if __name__ == "__main__":
    app.run(debug=True)
