def make_training_parameters(**kwargs):
    import os
    import math
    from atlas.configuration import config, constants
    from atlas.configuration import data_source as ds
    from atlas.data.image_provider import CHANNEL_HISTOLOGY
    from atlas.experiments import get_slices_in_range
    from atlas.torch.optimizers import WeightDecaySelector

    brain = __BRAIN__
    labels = ["bg", "cor", "wm", ] + __LABELS__
    sections = __SECTIONS__
    train_slices = [(brain, s) for s in sections]
    validation_sections = get_slices_in_range(start=min(sections),
            stop=max(sections),
            brain=brain,
            pattern=constants.original_path,
            step=1)
    congruent_slices = {key: train_slices for key in set(validation_sections)}

    # Apply weight decay
    param_selectors = [WeightDecaySelector(), WeightDecaySelector(inverse=True)]
    weight_decay = [0.0001, 0.0]
    learning_rate_params = config.LearningRateParams(learning_rate=0.01,
                                                     lr_policy="piecewise",
                                                     piecewise_steps=[1000, 1400, 1800, 2200, 2600],
                                                     piecewise_values=[0.01*(0.5 ** i) for i in range(1, 6)])
    learning_rate_params = config.LearningRateParamsPerParamGroup(learning_rate_params, learning_rate_params)

    train_params = config.TrainParams(
            backend="torch",
            mode="fast_segmentation",
            iterations=3000,
            test_interval=None,
            snapshot=100,
            data_format="NCHW",
            class_weights="$CLASS_WEIGHTS",
            save_best_weights=False,
            input_output_mapping=ds.InputOutputMapping(
                # Multi scale input
                input=("gray2", "gray16"),
                # Labels for supervised training
                output="labels"
                ),
            num_classes=len(labels),
            distributed_learning_rate_scaling=True,
            optimizer_params=config.OptimizerParams(
                # Note: Learning rate will be scaled by number of GPUs (most often 4)
                learning_rate_params=learning_rate_params,
                param_group_selectors=param_selectors,
                weight_decay=weight_decay,
            ),
                network_params=config.NetworkParams(
                        make_network=make_network,
                        finetune="/p/project/cjinm16/schiffer1/experiments/models/self_supervised/self_supervised_pretrained/model.hdf5",
                        finetune_mapping = "/p/project/cjinm16/schiffer1/experiments/models/keras2torch/self_supervised_pretrained_to_multi_scale_unet.json",
                    finetune_file_key=None,
                ),
            iterator_params={
                config.TRAIN_SPLIT: config.IteratorParams(
                    batch_size=12,
                    data_sources=[ds.ImageDataSource(name="gray2", size=2025, spacing=2, channels=CHANNEL_HISTOLOGY, data_format="NCHW"),
                        ds.ImageDataSource(name="gray16", size=628, spacing=16, channels=CHANNEL_HISTOLOGY, data_format="NCHW"),
                        ds.LabelDataSource(name="labels", size=68, spacing=16, labels=labels)],
                    multi_threaded_communications=11,
                    image_provider_params=config.ImageProviderParams(
                        transformation_parameters=config.TransformationParams(
                            normalize_orientation=True,
                            pre_registration_base_slice=sections[0],
                            brightness_augmentation=True,
                            brightness_augmentation_range=(-0.2, +0.2),
                            contrast_augmentation=True,
                            contrast_augmentation_range=(0.9, 1.1),
                            gamma_contrast_augmentation=True,
                            gamma_contrast_augmentation_range=(0.8, 1.214),
                            random_rotation=True,
                            random_rotation_range=(-math.pi / 4, +math.pi / 4)
                            ),
                        ),
                    sampler_params=config.SamplerParams(
                        sampler_mode="weighted",
                        weighted_sampler_probability_weights={label: 1 / len(labels) for label in labels},
                        weighted_sampler_labels=labels,
                        area_provider_params=config.AreaProviderParams(
                            area_mode="distance",
                            distance_area_provider_radius=2500,
                            distance_area_provider_radius_spacing=2,
                            distance_area_provider_radius_per_section=True,
                            congruent_enlargement_factor=1.05,
                            labels=labels,
                            congruent_slices=congruent_slices,
                            slices=train_slices,
                            use_pre_registration_for_congruent_sections=True,
                            congruent_mode="match",
                            ),
                        ),
                    ),
                config.VAL_SPLIT: config.IteratorParams(
                    batch_size=3,
                    data_sources=[ds.ImageDataSource(name="gray2", size=8169, spacing=2, channels=CHANNEL_HISTOLOGY, data_format="NCHW"),
                        ds.ImageDataSource(name="gray16", size=1396, spacing=16, channels=CHANNEL_HISTOLOGY, data_format="NCHW"),
                        ds.LabelDataSource(name="labels", size=68, spacing=16, labels=labels, create=False)],
                    multi_threaded_communications=11,
                    image_provider_params=config.ImageProviderParams(
                        transformation_parameters=config.TransformationParams(
                            normalize_orientation=True,
                            ),
                        ),
                    sampler_params=config.SamplerParams(
                        sampler_mode="grid",
                        grid_sampler_output_format=config.OutputFormat(size=836, spacing=16),
                        image_provider_params=config.ImageProviderParams(
                            transformation_parameters=config.TransformationParams(
                                normalize_orientation=True,
                                ),
                            ),
                        area_provider_params=config.AreaProviderParams(
                            area_mode="distance",
                            distance_area_provider_radius=2500,
                            distance_area_provider_radius_spacing=2,
                            distance_area_provider_radius_per_section=True,
                            congruent_enlargement_factor=1.05,
                            labels=labels,
                            slices=validation_sections,
                            congruent_slices=congruent_slices,
                            use_pre_registration_for_congruent_sections=True,
                            congruent_mode="match",
                            )
                        )
                    ),
                }
    )

    return train_params


def make_network(**kwargs):
    from atlas.torch.models.segmentation.unet_multiscale import UNet
    from atlas.torch.models.segmentation.unet_parts import UNetDownPathParams, UNetBottomPathParams, UNetUpPathParams

    input_shape = kwargs["input_shape"]
    input_spacing = kwargs["input_spacing"]
    num_classes = kwargs["num_classes"]

    input_channels = [s[1] for s in input_shape]
    model = UNet(input_channels=input_channels,
            input_spacings=input_spacing,
            num_classes=num_classes,
            down_params=[
                UNetDownPathParams(filters=[16, 32, 64, 64, 128],
                    kernel_size=[(5, 3), 3, 3, 3, 3],
                    stride=[(4, 1), 1, 1, 1, 1]),
                UNetDownPathParams(filters=[16, 32, 64, 64, 128, ],
                    dilation=[1, 2, 2, 2, 2])
                ],
            bottom_params=[
                UNetBottomPathParams(filters=[128, ]),
                UNetBottomPathParams(filters=[128, ], dilation=2)
                ],
            up_params=UNetUpPathParams(filters=[128, 64, 64, 32],
                upsampling_mode="upsampling"))
    return model

